from pathlib import Path
from typing import Union
from types import ModuleType
import os
import sys
import importlib


def import_module_from_path(module_path: Union[str, Path]) -> ModuleType:
    if isinstance(module_path, Path):
        module_path = str(module_path)
    module_name = os.path.basename(module_path)
    if os.path.isfile(module_path):
        sp = os.path.splitext(module_path)
        module_name = sp[0]

    if os.path.isfile(module_path):
        module_spec = importlib.util.spec_from_file_location(module_name, module_path)
        # module_dir = os.path.split(module_path)[0]
    else:
        module_spec = importlib.util.spec_from_file_location(
            module_name, os.path.join(module_path, "__init__.py")
        )
        # module_dir = module_path

    module = importlib.util.module_from_spec(module_spec)
    sys.modules[module_name] = module
    module_spec.loader.exec_module(module)
    return module



from .oneflow_exec_mode import oneflow_exec_mode, oneflow_exec_mode_enabled
from .env_var import (
    parse_boolean_from_env,
    set_boolean_env_var,
    parse_integer_from_env,
    set_integer_env_var,
)
from .model_inplace_assign import TensorInplaceAssign
from .version_util import (
    get_support_message,
    is_quantization_enabled,
    is_community_version,
)



def online_quantize_model(
    model, input_args, input_kwargs,
    seed=1, inplace=True,
    module_selector=lambda x: x,
    quant_config = None,
    calibration_info=None,
):
    """Optimize the quantization pipeline.

    Returns:
        tuple: A tuple containing the quantized model and the quantization
        status.
    """

    from onediff_quant.quantization import (
        OnlineQuantModule,
        create_quantization_calculator,
    )
    if getattr(quant_config, "quantization_calculator", None):
        calculator = quant_config.quantization_calculator
    else:
        calculator = create_quantization_calculator(
            model, quant_config, module_selector, seed,
            calibration_info=calibration_info,
        )
    module = OnlineQuantModule(calculator, False, inplace=inplace)

    quantized_model,  info = module.quantize_with_calibration(
        *input_args, **input_kwargs
    )
    status = module.collect_quantization_status(model, info)

    return quantized_model, status


def quantize_and_deploy_wrapper(func):
    def wrapper(self: "DeployableModule", *args, **kwargs):
        torch_model = self._torch_module
        quant_config = self._deployable_module_quant_config
        if quant_config:
            torch_model, _ = online_quantize_model(
                torch_model, args, kwargs,
                module_selector=lambda x: x,
                quant_config=quant_config,
                inplace=True,
            )
            self._deployable_module_quant_config = None 
        output = func(self, *args, **kwargs)
        return output
    return wrapper

       


from functools import wraps

from ..transform.builtin_transform import torch2oflow
from ..transform.manager import transform_mgr
from ..utils.log_utils import logger
from .dual_module import DualModule


@torch2oflow.register
def _(mod: DualModule, verbose=False):
    return torch2oflow(mod._torch_module, verbose)


def handle_deployable_exception(func):
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if transform_mgr.debug_mode:
            return func(self, *args, **kwargs)
        else:
            try:
                return func(self, *args, **kwargs)
            except Exception as e:
                logger.error(f"Exception in {func.__name__}: {e=}")
                logger.warning("Recompile oneflow module ...")
                del self._deployable_module_model.oneflow_module
                self._deployable_module_dpl_graph = None
                return func(self, *args, **kwargs)

    return wrapper


def get_mixed_dual_module(module_cls):
    if issubclass(module_cls, DualModule) and "MixedDualModule" in module_cls.__name__:
        return module_cls

    class MixedDualModule(DualModule, module_cls):
        def __init__(self, torch_module, oneflow_module):
            while isinstance(torch_module, DualModule):
                torch_module = torch_module._torch_module
            DualModule.__init__(self, torch_module, oneflow_module)

        def _get_name(self) -> str:
            return f"{self.__class__.__name__}(of {module_cls.__name__})"

    return MixedDualModule


# Return a OneflowDeployableModule that using module_cls as it's parent class.
def get_mixed_deployable_module(module_cls):
    from .deployable_module import OneflowDeployableModule

    class MixedOneflowDeployableModule(OneflowDeployableModule, module_cls):
        def __init__(self, torch_module, oneflow_module, dynamic=True, options=None):
            OneflowDeployableModule.__init__(
                self, torch_module, oneflow_module, dynamic, options
            )
            self._is_raw_deployable_module = False

        @classmethod
        def from_existing(cls, existing_module, dynamic=True, options=None):
            torch_module = existing_module._deployable_module_model._torch_module
            oneflow_module = existing_module._deployable_module_model._oneflow_module
            instance = cls(torch_module, oneflow_module, dynamic, options)
            instance._deployable_module_dpl_graph = None
            if hasattr(existing_module, "_deployable_module_dpl_graph"):
                instance._deployable_module_dpl_graph = (
                    existing_module._deployable_module_dpl_graph
                )
            return instance

        def _get_name(self):
            return f"{self.__class__.__name__}(of {module_cls.__name__})"

    return MixedOneflowDeployableModule


def get_oneflow_graph(model, size=9, dynamic_graph=True):
    from .graph import OneflowGraph

    g = OneflowGraph(model)
    g._dynamic_input_graph_cache.set_cache_size(size)
    g._dynamic_input_graph_cache.enable_shared(dynamic_graph)
    return g
